#include "mainwindow.h"
#include "ui_mainwindow.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MainWindow::openNarc(){
    this->ui->narcEntryIds->clear();
    QFileDialog dialog(this);
    dialog.setFileMode(QFileDialog::ExistingFile);
    QStringList filters;
    filters<<"narc (*.narc)";
    filters<<"All (*)";
    dialog.setNameFilters(filters);
    QStringList fileNames;
    if(dialog.exec())
    {
        fileNames = dialog.selectedFiles();
        this->narcName=fileNames[0];
        QFile file(fileNames[0]);
        this->fileName=file.fileName();
        file.open(QIODevice::ReadOnly);
        this->narc=Narc(file,0);
        this->ui->narcEntryIds->clear();
        for(int i=0;i<this->narc.numEntries;i++)
        {
            this->ui->narcEntryIds->addItem(QString::number(i)+" - "+QString::number(this->narc.endOffsets[i]-this->narc.startOffsets[i],16));
        }
        this->report("Loaded "+this->narcName);
    }
}

void MainWindow::writeToNarc()
{
    QFile file(this->narcName);
    file.open(QIODevice::WriteOnly);
    narc.writeNarc(file,0);
    this->report("Finished writing "+this->narcName);
}

void MainWindow::typeChanged()
{
    int m=this->ui->narcEntryIds->currentIndex();
    int i=this->ui->typeIds->currentIndex();
    QStringList strings=this->getStrings(m,i);
    this->ui->text->setPlainText(strings.join("\n"));
}

void MainWindow::loadEntryFromNarc()
{
    if(this->ui->narcEntryIds->count()>0)
    {
        this->ui->typeIds->clear();
        this->ui->typeIds->addItem("Hiragana");
        if(this->narc.fileData[this->ui->narcEntryIds->currentIndex()][0]==2)
        {
            this->ui->typeIds->addItem("Kanji");
        }
    }
}

void MainWindow::saveEntry(){
    this->saveEntry(this->ui->text->toPlainText(),this->ui->narcEntryIds->currentIndex(),this->ui->typeIds->currentIndex());
}
void MainWindow::saveEntry(QString string,int m, int i){
    //m = narc Entry, i=type
    QStringList text;
    if(string.contains("\r\n"))
    {
        text=string.split("\r\n",QString::KeepEmptyParts);
    }
    else if(string.contains('\r'))
    {
        text=string.split('\r',QString::KeepEmptyParts);
    }
    else
    {
        text=string.split('\n',QString::KeepEmptyParts);
    }
    QString numString="%1";
    QStringList oText=this->getStrings(m,i);
    DataStream ds(this->narc.fileData[m]);
    quint16 numEntries,numSections;
    quint32 unk1;
    quint32 sizeSections[]={0,0,0};
    quint32 sectionOffset[]={0,0,0};
    quint32 newsizeSections[]={0,0,0};
    quint32 newsectionOffset[]={0,0,0};
    ds>>numSections>>numEntries>>sizeSections[0]>>unk1;
    if(i<numSections)
    {
        if(text.count()<numEntries)
        {
            numString="%1";
            this->report("Skipping "+numString.arg(i,4,10,QChar('0'))+"/"+numString.arg(m,4,10,QChar('0'))+" due to too few lines");
        }
        else
        {
            QVector<quint8> newEntry=this->makeSection(text,numEntries);
            if(i==0)
            {
                for(int z=0;z<numSections;z++)
                {
                    ds>>sectionOffset[z];
                }
                for(int z=0;z<numSections;z++)
                {
                    ds.pos=sectionOffset[z];
                    ds>>sizeSections[z];
                }
                newsizeSections[0]=newEntry.size();
                ds.pos=4;
                ds<<newsizeSections[0];
                if(numSections==2)
                {
                    newsectionOffset[1]=newsizeSections[0]+sectionOffset[0];
                    ds.pos=0x10;
                    ds<<newsectionOffset[1];
                }
                ds.close(this->narc.fileData[m]);
                QVector<quint8> section2;
                if(numSections==2)
                {
                    section2=this->narc.fileData[m].mid(sectionOffset[1],sizeSections[1]);
                }
                this->narc.fileData[m].remove(sectionOffset[0],sizeSections[0]+sizeSections[1]);
                this->narc.fileData[m]+=newEntry;
                if(numSections==2)
                {
                    this->narc.fileData[m]+=section2;
                }
            }
            else if(numSections==2)
            {
                DataStream ds(this->narc.fileData[m]);
                quint32 sizeSections[]={0,0,0};
                quint32 sectionOffset[]={0,0,0};
                ds.pos=2;
                quint16 numEntries;
                ds>>numEntries;
                ds.pos=0xC;
                for(int z=0;z<numSections;z++)
                {
                    ds>>sectionOffset[z];
                }
                for(int z=0;z<numSections;z++)
                {
                    ds.pos=sectionOffset[z];
                    ds>>sizeSections[z];
                }
                ds.close(this->narc.fileData[m]);
                this->narc.fileData[m].remove(sectionOffset[1],sizeSections[1]);
                this->narc.fileData[m]+=newEntry;
            }
        }
    }
}

QVector<quint8> MainWindow::makeSection(QStringList strings,int numEntries)
{
    QList<QList<quint16> > data;
    quint32 size=0;
    quint32 offset=4+8*numEntries;
    quint16 charCount,unk1=0x100;
    for(int i=0;i<numEntries;i++)
    {
        data.append(this->parseString(strings[i],i));
        size+=(data[i].length()*2);
    }
    if(size%4==2)
    {
        size+=2;
        quint16 tmpKey=this->keys[numEntries-1];
        //quint16 tmpKey=0x7C89;
        for(int i=0;i<data[numEntries-1].length();i++)
        {
            tmpKey=((tmpKey << 3) | (tmpKey>> 13)) & 0xffff;
        }
        data[numEntries-1].append(0xFFFF^tmpKey);
    }
    size+=offset;
    QVector<quint8> section;
    section.fill(0,size);
    DataStream ds(section);
    ds.pos=0;
    ds<<size;
    for(int i=0;i<numEntries;i++)
    {
        charCount=data[i].length();
        ds<<offset<<charCount<<this->unknowns[i];
        offset+=charCount*2;
    }
    for(int i=0;i<numEntries;i++)
    {
        for(int j=0;j<data[i].length();j++)
        {
            ds<<data[i][j];
        }
    }
    ds.close(section);
    return section;
}

QList<quint16> MainWindow::parseString(QString string,int entry_id)
{
    QList<quint16> chars;
    QString tmp;
    bool ok;
    for(int i=0;i<string.length();i++)
    {
        if(string[i]!='\\')
        {
            chars.append(string[i].unicode());
        }
        else
        {
            if((i+2<string.length()) && string[i+2]=='{')
            {
                chars.append(string[i].unicode());
            }
            else
            {
                tmp="";
                for(int j=0;j<4;j++)
                {
                    tmp+=string[i+j+2];
                }
                i+=5;
                chars.append(tmp.toUShort(&ok,16));
            }
        }
    }
    chars.append(0xffff);
    quint16 key=this->keys[entry_id];
    //quint16 key=0x7C89;
    for(int i=0;i<chars.length();i++)
    {
        chars[i]^=key;
        key=((key << 3) | (key >> 13)) & 0xffff;
    }
    return chars;
}

void MainWindow::exportToFiles()
{
    QString dir_path="pptxt_files"+this->fileName+"/";
    QString narcId,typeId;
    QFileDialog dialog(this);
    dialog.setFileMode(QFileDialog::DirectoryOnly);
    QStringList fileNames;
    if(dialog.exec())
    {
        fileNames = dialog.selectedFiles();
        if(fileNames[0][fileNames[0].length()-1]!='/')
        {
            fileNames[0]+="/";
        }
        this->ui->progressBar->setRange(0,this->narc.numEntries-1);
        this->ui->label_2->setText("Exporting: ");
        QDir dir(fileNames[0]);
        if(dir.cd("0000")&&dir.cd("../0001"))
        {
            this->report("Exporting to directory "+fileNames[0]);
            for(int i=0;i<this->narc.numEntries;i++)
            {
                this->ui->progressBar->setValue(i);
                for(int j=0;j<2;j++)
                {
                    QString string=this->getStrings(i,j).join("\n");
                    if(string!="")
                    {
                        narcId="%1";
                        typeId="%1";
                        QFile file(fileNames[0]+typeId.arg(j,4,10,QChar('0'))+"/"+narcId.arg(i,4,10,QChar('0'))+".txt");
                        file.open(QIODevice::WriteOnly);
                        QTextStream ts(&file);
                        ts.setCodec("UTF-8");
                        ts<<string;
                        file.close();
                    }
                }
            }
            this->ui->label_2->setText("Finished Exporting!");
        }
        else
        {
            this->report("Failed to export. Either 0000 or 0001 does not exist in "+fileNames[0]);
        }
    }
}
void MainWindow::exportToFile()
{
    QString narcId,typeId;
    QFileDialog dialog(this);
    dialog.setFileMode(QFileDialog::AnyFile);
    QStringList fileNames;
    if(dialog.exec())
    {
        fileNames = dialog.selectedFiles();
        this->ui->progressBar->setRange(0,this->narc.numEntries-1);
        this->ui->label_2->setText("Exporting: ");
        this->report("Exporting to file "+fileNames[0]);
        QFile file(fileNames[0]);
        if(file.open(QIODevice::WriteOnly))
        {
            QTextStream ts(&file);
            ts.setCodec("UTF-8");
            for(int i=0;i<this->narc.numEntries;i++)
            {
                ts << "****Entry "+QString::number(i)+":\n";
                this->ui->progressBar->setValue(i);
                for(int j=0;j<2;j++)
                {
                    ts << "****Entry "+QString::number(i)+": Section "+QString::number(j)+":\n";
                    QString string=this->getStrings(i,j).join("\n");
                    if(string!="")
                    {
                        narcId="%1";
                        typeId="%1";
                        ts<<string;
                    }
                }
            }
            file.close();
            this->ui->label_2->setText("Finished Exporting!");
        }
        else
        {
            this->report("Failed to export to file "+fileNames[0]);
        }
    }
}
void MainWindow::importFromFiles()
{
    QString dir_path="pptxt_files"+this->fileName+"/";
    QString narcId,typeId;
    QFileDialog dialog(this);
    dialog.setFileMode(QFileDialog::DirectoryOnly);
    QStringList fileNames;
    QString tmpText;
    if(dialog.exec())
    {
        fileNames = dialog.selectedFiles();
        if(fileNames[0][fileNames[0].length()-1]!='/')
        {
            fileNames[0]+="/";
        }
        this->ui->progressBar->setRange(0,this->narc.numEntries-1);
        this->ui->label_2->setText("Importing: ");
        QDir dir(fileNames[0]);
        if(dir.cd("0000")&&dir.cd("../0001"))
        {
            this->report("Importing from directory "+fileNames[0]);
            for(int i=0;i<this->narc.numEntries;i++)
            {
                this->ui->progressBar->setValue(i);
                for(int j=0;j<2;j++)
                {
                    narcId="%1";
                    typeId="%1";
                    this->ui->typeIds->setCurrentIndex(j);
                    QFile file(fileNames[0]+typeId.arg(j,4,10,QChar('0'))+"/"+narcId.arg(i,4,10,QChar('0'))+".txt");
                    if(file.open(QIODevice::ReadOnly /*| QIODevice::Text*/))
                    {
                        QTextStream ts(&file);
                        ts.setCodec("UTF-8");
                        this->saveEntry(ts.readAll(),i,j);
                        file.close();
                    }
                    else
                    {
                        narcId="%1";
                        typeId="%1";
                        this->report("Failed to import from file: "+fileNames[0]+typeId.arg(j,4,10,QChar('0'))+"/"+narcId.arg(i,4,10,QChar('0'))+".txt");
                    }
                }
            }
            this->ui->label_2->setText("Finished Importing!");
            this->report("Finished importing from directory "+fileNames[0]);
        }
        else
        {
            this->report("Failed to import. Either 0000 or 0001 does not exist in "+fileNames[0]);
        }
    }
}

QStringList MainWindow::getStrings(int m,int i)
{
    QStringList strings;
    this->keys.clear();
    this->unknowns.clear();
    if(i>=0&& m>=0)
    {
        DataStream ds(this->narc.fileData[m]);
        ds.pos=0;
        quint16 numSections,numEntries,tmpCharCount,tmpUnknown,tmpChar;
        quint32 unk1,tmpOffset;
        quint32 sizeSections[]={0,0,0};
        quint32 sectionOffset[]={0,0,0};
        QMap<int,QList<quint32> > tableOffsets;
        QMap<int,QList<quint16> > characterCount;
        QMap<int,QList<quint16> > unknown;
        QMap<int,QList<QList<quint16> > > encText;
        QMap<int,QList<QStringList> > decText;
        QString string;
        quint16 key;

        ds>>numSections>>numEntries>>sizeSections[0]>>unk1;
        if(numSections>i)
        {
            for(int z=0;z<numSections;z++)
            {
                ds>>sectionOffset[z];
            }
            ds.pos=sectionOffset[i];
            ds>>sizeSections[i];
            for(int j=0;j<numEntries;j++)
            {
                ds>>tmpOffset>>tmpCharCount>>tmpUnknown;
                tableOffsets[i].append(tmpOffset);
                characterCount[i].append(tmpCharCount);
                unknown[i].append(tmpUnknown);
                this->unknowns.append(tmpUnknown);
            }
            for(int j=0;j<numEntries;j++)
            {
                QList<quint16> tmpEncChars;
                ds.pos=sectionOffset[i]+tableOffsets[i][j];
                for(int k=0;k<characterCount[i][j];k++)
                {
                    ds>>tmpChar;
                    tmpEncChars.append(tmpChar);
                }
                encText[i].append(tmpEncChars);
                key=encText[i][j][characterCount[i][j]-1]^0xFFFF;
                for(int k=characterCount[i][j]-1;k>=0;k--)
                {
                    encText[i][j][k]^=key;
                    if(k==0)
                    {
                        this->keys.append(key);
                    }
                    key=((key >> 3) | (key << 13)) & 0xffff;
                }

                QStringList chars;
                string="";
                for(int k=0;k<characterCount[i][j];k++)
                {
                    if(encText[i][j][k]==0xFFFF)
                    {
                        //chars.append("\\xffff");
                    }
                    else
                    {
                        //if((encText[i][j][k]<0xF000&&encText[i][j][k]>3) || (encText[i][j][k]>=0xFF00 && ((encText[i][j][k]&0xFF)>1&&(encText[i][j][k]&0xFF)<0xF0)))
                        if(encText[i][j][k]>20 && encText[i][j][k]<=0xFFF0&& encText[i][j][k]!=0xF000&& QChar::unicodeVersion(encText[i][j][k])!=QChar::Unicode_Unassigned)
                        {
                            chars.append(QString(QChar(encText[i][j][k])));
                        }
                        else
                        {
                            QString num=QString::number(encText[i][j][k],16);
                            int length=num.length();
                            for(int l=0;l<4-length;l++)
                            {
                                num="0"+num;
                            }
                            chars.append("\\x"+num);
                        }
                        string+=chars[k];
                    }
                }
                strings.append(string);
                decText[i].append(chars);
            }
        }
    }
    return strings;
}

void MainWindow::report(QString string)
{
    this->ui->report->setPlainText(this->ui->report->toPlainText()+string+"\n");
}
